/*
//
//  talloc.h
//  neblo
//  Thread Safe Memory Allocation for the GNU Libc
//  Created by Paolo Gianrossi on 2006-11-13.
//  Copyright (c) 2006 Paolo Gianrossi, under Gnu GPL
//
*/

#ifndef _TALLOC_H_
#define _TALLOC_H_

/* Include config file if it exists (for use with autotools)*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* malloc mutex definition */
#ifdef HAVE_LIBPTHREAD
#include <pthread.h>
extern pthread_mutex_t malloc_lock;
#endif

#include <stdlib.h>
#include <sys/types.h>

/* 
Actually malloc.h is a GNU Extension.. All this won't work on other
implementations of libc, but that is usually useless, since most
implementations of libc (except GNU) have already thread-safe implementations
of malloc and friends. 
*/
#ifdef HAVE_MALLOC_H
#include <malloc.h>

/*Also, we need to be sure that we have pthreads. */
#ifdef HAVE_LIBPTHREAD

/* Placeholders for original hooks */
static void * (*old_malloc_hook) (size_t, const void *);
static void   (*old_free_hook) (void*, const void *);
static void * (*old_realloc_hook)(void*, size_t, const void *);
static void * (*old_memalign_hook)(size_t, size_t, const void *);

/*
	SAVEANDSET will save current malloc hooks and set our own (in the form
	of myprefix_function_hook)
*/
#define SAVEANDSET(myprefix) { 				\
	old_malloc_hook   = __malloc_hook;		\
	old_free_hook     = __free_hook;		\
	old_realloc_hook  = __realloc_hook;		\
	old_memalign_hook = __memalign_hook;		\
							\
	__malloc_hook     = myprefix##_malloc_hook;	\
	__free_hook       = myprefix##_free_hook;	\
	__realloc_hook    = myprefix##_realloc_hook;	\
	__memalign_hook   = myprefix##_memalign_hook;	\
}

/*
	RESTOREOLD will set old hooks back
*/
#define RESTOREOLD	{				\
	__malloc_hook     = old_malloc_hook;		\
	__free_hook       = old_free_hook;		\
	__realloc_hook    = old_realloc_hook;		\
	__memalign_hook   = old_memalign_hook;		\
}


/* Prototypes for our hooks.  */
static void pthread_init_hook (void);
static void *pthread_malloc_hook (size_t, const void *);
static void pthread_free_hook (void*, const void *);
static void *pthread_realloc_hook(void*, size_t, const void *);
static void *pthread_memalign_hook(size_t, size_t, const void *);


/* Override initializing hook from the C library. */
static void (*__malloc_initialize_hook) (void) = pthread_init_hook;

/* Initialization function will just set our own hooks */
static void pthread_init_hook (void) { SAVEANDSET(pthread); }

/* 
	All our hooks basically just do the same thing: lock the mutex, call
	the original function, do usual housekeeping and unlock the mutex.
*/

static void *
pthread_malloc_hook (size_t size, const void *caller)
{
	void *result;
	/*lock mutex */
	pthread_mutex_lock(&malloc_lock);
	RESTOREOLD;
	/* Call recursively */
	result = malloc (size);
	SAVEANDSET(pthread);	
	/*unlock mutex*/
	pthread_mutex_unlock(&malloc_lock);
	return result;
}
     
static void *
pthread_realloc_hook (void *ptr, size_t size, const void *caller)
{
	void *result;
	/*lock mutex */
	pthread_mutex_lock(&malloc_lock);
	RESTOREOLD;
	/* Call recursively */
	result = realloc (ptr, size);
	SAVEANDSET(pthread);	
	/*unlock mutex*/
	pthread_mutex_unlock(&malloc_lock);
	return result;
}

static void *
pthread_memalign_hook (size_t alignment, size_t size, const void *caller)
{
	void *result;
	/*lock mutex */
	pthread_mutex_lock(&malloc_lock);
	RESTOREOLD;
	/* Call recursively */
	result = memalign (alignment, size);
	SAVEANDSET(pthread);	
	/*unlock mutex*/
	pthread_mutex_unlock(&malloc_lock);
	return result;
}

static void
pthread_free_hook (void *ptr, const void *caller)
{
	/* lock mutex */
	pthread_mutex_lock(&malloc_lock);
	RESTOREOLD;
    /* Call recursively */
	free (ptr);
	SAVEANDSET(pthread);
	/*unlock mutex*/
	pthread_mutex_unlock(&malloc_lock);
}

#endif	/* PTHREADS_MEMORY */

#endif	/* HAVE_MALLOC_H */

/* 
We also define the usual functions and macros, xmalloc, xrealloc and
xcalloc.  
*/

static __inline__ void* _talloc_xmalloc(size_t size)
{
	void* result;
	result=malloc(size);
	if(!result){
		perror("Mallocating memory -");
		exit(EXIT_FAILURE);
	}
	return result;
}

static __inline__ void* _talloc_xcalloc(size_t size, size_t number)
{
	void* result;
	result = calloc(size, number);
	if(!result){
		perror("Callocating memory -");
		exit(EXIT_FAILURE);
	}
	return result;
}

static __inline__ void* _talloc_xrealloc(void *ptr, size_t size)
{
	void* result;
	result=realloc(ptr, size);
	if(!result){
		perror("Rellocating memory -");
		exit(EXIT_FAILURE);
	}
	return result;
}

#define xmalloc _talloc_xmalloc
#define xcalloc _talloc_xcalloc
#define xrealloc _talloc_xrealloc

#endif /* _TALLOC_H_ */
